@font-face {
  font-family: "Inter";
  src: url("./assets/Inter-VariableFont_slnt,wght.ttf") format("truetype");
  font-weight: 1 100 200 300 400 500 600 700 800 900 1000;
  font-style: normal italic;
}

body,
html {
  width: 100%;
  height: 100%;
  margin: 0;
  padding: 0;
  overflow: hidden;
}

body {
  font-family:
    "Inter",
    -apple-system,
    BlinkMacSystemFont,
    "Segoe UI",
    "Roboto",
    "Oxygen",
    "Ubuntu",
    "Cantarell",
    "Fira Sans",
    "Droid Sans",
    "Helvetica Neue",
    sans-serif;
  -webkit-font-smoothing: antialiased;
  -moz-osx-font-smoothing: grayscale;
}

code {
  font-family: source-code-pro, Menlo, Monaco, Consolas, "Courier New",
    monospace;
}

#root {
  width: 100%;
  height: 100%;
  overflow: hidden;
}

/* Styling for threejs stats panel. Switches position: fixed to position:
 * absolute, to respect parent + for better multi-pane layout. */
.stats-panel {
  position: absolute !important;
}

/* Styling for color chips in markdown */
.gfm-color-chip {
  margin-left: 0.125rem;
  display: inline-block;
  height: 0.625rem;
  width: 0.625rem;
  border-radius: 9999px;
  border: 1px solid gray;
  transform: TranslateY(0.07em);
}
