// Automatically generated file - do not edit manually.
// This is synchronized with the Python package version in viser/__init__.py.
export const VISER_VERSION = "1.0.24";

// GitHub contributors for the viser project.
export interface Contributor {
  login: string;
  html_url: string;
}

export const GITHUB_CONTRIBUTORS: Contributor[] = [
  {
    login: "brentyi",
    html_url: "https://github.com/brentyi",
  },
  {
    login: "kerrj",
    html_url: "https://github.com/kerrj",
  },
  {
    login: "chungmin99",
    html_url: "https://github.com/chungmin99",
  },
  {
    login: "tancik",
    html_url: "https://github.com/tancik",
  },
  {
    login: "jonahbedouch",
    html_url: "https://github.com/jonahbedouch",
  },
  {
    login: "hangg7",
    html_url: "https://github.com/hangg7",
  },
  {
    login: "BrianSantoso",
    html_url: "https://github.com/BrianSantoso",
  },
  {
    login: "origamiman72",
    html_url: "https://github.com/origamiman72",
  },
  {
    login: "ginazhouhuiwu",
    html_url: "https://github.com/ginazhouhuiwu",
  },
  {
    login: "sea-bass",
    html_url: "https://github.com/sea-bass",
  },
  {
    login: "jkulhanek",
    html_url: "https://github.com/jkulhanek",
  },
  {
    login: "ethanweber",
    html_url: "https://github.com/ethanweber",
  },
  {
    login: "zerolover",
    html_url: "https://github.com/zerolover",
  },
  {
    login: "henrynoyes",
    html_url: "https://github.com/henrynoyes",
  },
  {
    login: "beckyfeng08",
    html_url: "https://github.com/beckyfeng08",
  },
  {
    login: "vye16",
    html_url: "https://github.com/vye16",
  },
  {
    login: "david-dorf",
    html_url: "https://github.com/david-dorf",
  },
  {
    login: "AdamRashid96",
    html_url: "https://github.com/AdamRashid96",
  },
  {
    login: "slecleach",
    html_url: "https://github.com/slecleach",
  },
  {
    login: "pritzza",
    html_url: "https://github.com/pritzza",
  },
  {
    login: "yzslab",
    html_url: "https://github.com/yzslab",
  },
  {
    login: "kikislater",
    html_url: "https://github.com/kikislater",
  },
  {
    login: "simonbethke",
    html_url: "https://github.com/simonbethke",
  },
  {
    login: "swnakamura",
    html_url: "https://github.com/swnakamura",
  },
  {
    login: "rowoflo",
    html_url: "https://github.com/rowoflo",
  },
  {
    login: "MosesEbere",
    html_url: "https://github.com/MosesEbere",
  },
  {
    login: "Mitia-RH",
    html_url: "https://github.com/Mitia-RH",
  },
  {
    login: "nlml",
    html_url: "https://github.com/nlml",
  },
  {
    login: "kevinzakka",
    html_url: "https://github.com/kevinzakka",
  },
  {
    login: "kdexd",
    html_url: "https://github.com/kdexd",
  },
  {
    login: "ojh6404",
    html_url: "https://github.com/ojh6404",
  },
  {
    login: "heiwang1997",
    html_url: "https://github.com/heiwang1997",
  },
  {
    login: "hongsukchoi",
    html_url: "https://github.com/hongsukchoi",
  },
  {
    login: "mcallisterdavid",
    html_url: "https://github.com/mcallisterdavid",
  },
  {
    login: "cvachha",
    html_url: "https://github.com/cvachha",
  },
  {
    login: "ArthurAllshire",
    html_url: "https://github.com/ArthurAllshire",
  },
  {
    login: "lukehugh",
    html_url: "https://github.com/lukehugh",
  },
  {
    login: "escontra",
    html_url: "https://github.com/escontra",
  },
  {
    login: "alberthli",
    html_url: "https://github.com/alberthli",
  },
  {
    login: "atonderski",
    html_url: "https://github.com/atonderski",
  },
];
